/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>
#include <png.h>
#include <setjmp.h>
#include "common.h"
#include "../imgview.h"


static void png_errmsg(png_structp png_ptr, png_const_charp msg)
{
    message("PNG Library Error", (char *)msg, NULL);
}

static void png_warnmsg(png_structp png_ptr, png_const_charp msg)
{
    message("PNG Library Warning", (char *)msg, NULL);
}

XImage *readpng(int fd)
{
    FILE *f;
    XImage *volatile img = NULL;
    png_structp png_ptr = NULL;
    png_infop info_ptr = NULL;
    png_text *tcomments;
    int depth, ctype;
    int bilvl = 0;
    png_uint_32 w, h;
    double igamma;
    unsigned char *volatile buf = NULL, *bp;
    png_byte **volatile row_pointers = NULL;
    int i, rowbytes, channels;
    RGB bg, rgb, *rgbp;
    int x, y;
    unsigned char *g2rbuf = NULL;

    f = fdopen(fd, "r");

    if (! (png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
            png_errmsg, png_warnmsg)))
        goto err;
    if (! (info_ptr = png_create_info_struct(png_ptr)))
        goto err;

    if (setjmp(png_jmpbuf(png_ptr))) {
        if (img) { rmximage(img); img = NULL; }
        goto err;
    }

    png_init_io(png_ptr, f);
    png_read_info(png_ptr, info_ptr);

    png_get_IHDR(png_ptr, info_ptr, &w, &h, &depth, &ctype, NULL, NULL, NULL);

    {
        png_uint_32 res_x, res_y;
        int unit;
        char phys_str[80] = "";
        char gamma_str[80] = "";
        if (png_get_pHYs(png_ptr, info_ptr, &res_x, &res_y, &unit)) {
            if (unit == PNG_RESOLUTION_METER) {
                v.ippi[0] = (double)res_x * .0254;
                v.ippi[1] = (double)res_y * .0254;
            }
            if (res_x && res_y) {
                if (unit == PNG_RESOLUTION_METER)
                    v.par = (double)res_x / (double)res_y;
                else
                    v.par = (double)res_y / (double)res_x;
            }
            sprintf(phys_str, "%s: %g%s%g%s\n",
                    (unit == PNG_RESOLUTION_METER) ? "Resolution" :
                            "Pixel Aspect Ratio",
                    (unit == PNG_RESOLUTION_METER) ? v.ippi[0] :
                            (double)res_x,
                    (unit == PNG_RESOLUTION_METER) ? " × " : ":",
                    (unit == PNG_RESOLUTION_METER) ? v.ippi[1] :
                            (double)res_y,
                    (unit == PNG_RESOLUTION_METER) ? " pixels per inch" :
                            "");
        }
        if (png_get_gAMA(png_ptr, info_ptr, &igamma)) {
            sprintf(gamma_str, "\nGamma: %g", igamma);
        }
        sprintf(v.xinfostr, "%s"
                "Bits Per Sample: %d\n"
                "Color Type: %s"
                "%s",
                phys_str, depth, (ctype == 0) ? "Grayscale" :
                (ctype == 2) ? "RGB" : (ctype == 3) ? "Palette" :
                (ctype == 4) ? "Grayscale with Alpha" :
                (ctype == 6) ? "RGB with Alpha" : "Unknown",
                gamma_str);
    }

    if (depth == 1 && ctype == PNG_COLOR_TYPE_GRAY &&
            (! png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))) {
        bilvl++; v.black = 0; v.white = 1;
    }

    if (ctype == PNG_COLOR_TYPE_PALETTE) {
        if (depth == 1 &&
                (! png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))) {
            png_colorp pal;
            int numpal;
            if (png_get_PLTE(png_ptr, info_ptr, &pal, &numpal)) {
                v.black = v.white = -1;
                if (pal[0].red == 0 && pal[0].green == 0 &&
                        pal[0].blue == 0)
                    v.black = 0;
                else if (pal[0].red == 255 && pal[0].green == 255 &&
                        pal[0].blue == 255)
                    v.white = 0;
                if (pal[1].red == 0 && pal[1].green == 0 &&
                        pal[1].blue == 0)
                    v.black = 1;
                else if (pal[1].red == 255 && pal[1].green == 255 &&
                        pal[1].blue == 255)
                    v.white = 1;
                if (v.black != -1 && v.white != -1) bilvl++;
            }
        }
        if (! bilvl) png_set_expand(png_ptr);
    }

    if (! bilvl && v.hOutProfile) {
        char *name;
        int comp;
        if (png_get_iCCP(png_ptr, info_ptr, &name, &comp,
                    &v.iprof, (png_uint_32 *)&v.iproflen))
            v.hInProfile = cmsOpenProfileFromMem(v.iprof, v.iproflen);
        else
            v.hInProfile = ctype & PNG_COLOR_MASK_COLOR ? v.hRGBProfile :
                    v.hGrayProfile;
    }

    if (! bilvl) {
        if (ctype == PNG_COLOR_TYPE_GRAY)
            png_set_expand(png_ptr);
        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
            png_set_expand(png_ptr);
        if (depth == 16)
            png_set_strip_16(png_ptr);
        if (!v.hInProfile && png_get_gAMA(png_ptr, info_ptr, &igamma)) {
            char *sgamma_s = getenv("DISPLAY_GAMMA");
            double sgamma;
            if (sgamma_s) sgamma = atof(sgamma_s); else sgamma = 2.2;
            png_set_gamma(png_ptr, sgamma, igamma);
        }
        bg.r = v.tbgcolor_rgb.red >> 8;
        bg.g = v.tbgcolor_rgb.green >> 8;
        bg.b = v.tbgcolor_rgb.blue >> 8;
    }

    png_read_update_info(png_ptr, info_ptr);

    rowbytes = png_get_rowbytes(png_ptr, info_ptr);

    v.pad = 0;
    img = mkximage(w, h, bilvl ? 1 : xv.depth, &v.pad);

    row_pointers = (png_bytepp)malloc(h * sizeof(png_bytep));
    if (bilvl) {
        for (i = 0; i < h; i++)
            row_pointers[i] = (png_byte *)img->data + i * rowbytes;
    } else {
        buf = (unsigned char *)malloc(rowbytes * h);
        for (i = 0; i < h; i++)
            row_pointers[i] = buf + i * rowbytes;
    }

    png_read_image(png_ptr, row_pointers);
    if (bilvl) goto end;

    channels = png_get_channels(png_ptr, info_ptr);

    if (v.hInProfile && v.hOutProfile && !isoutprof()) {
        int nc = channels;
        v.hTransform = cmsCreateTransform(
                v.hInProfile,
                nc == 4 ? TYPE_RGBA_8 : nc == 3 ? TYPE_RGB_8 :
                        nc == 2 ? TYPE_GRAYA_8 : TYPE_GRAY_8,
                v.hOutProfile,
                nc == 2 || nc == 4 ? TYPE_RGBA_8 : TYPE_RGB_8,
                INTENT_PERCEPTUAL, 0);
        if (v.hTransform && nc <= 2)
            g2rbuf = (unsigned char *)malloc(w * (nc == 2 ? 4 : 3));
    }

    dp = (unsigned char *)img->data;
    idp = (unsigned int *)dp;

    for (y = 0; y < h; y++) {
        bp = row_pointers[y];
        int nc = channels;
        if (v.hTransform) {
            unsigned char *obp;
            if (nc <= 2) {
                obp = g2rbuf;
                if (nc == 2) {
                    unsigned char *p1 = bp;
                    unsigned char *p2 = g2rbuf;
                    for (i = 0; i < w; i++) {
                        *(p2 + 3) = *(p1 + 1);
                        p1 += 2; p2 += 4;
                    }
                }
            } else
                obp = bp;
            cmsDoTransform(v.hTransform, bp, obp, w);
            bp = obp;
            nc = (nc == 1) ? 3 : (nc == 2) ? 4 : nc;
        }
        if (nc == 4) {
            unsigned char a;
            for (x = 0; x < w; x++) {
                rgb.r = *bp++; rgb.g = *bp++; rgb.b = *bp++; a = *bp++;
                if (a == 0)
                    rgb = bg;
                else if (a != 255) {
                    png_composite(rgb.r, rgb.r, a, bg.r);
                    png_composite(rgb.g, rgb.g, a, bg.g);
                    png_composite(rgb.b, rgb.b, a, bg.b);
                }
                DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
            }
        } else if (nc == 3) {
            rgbp = (RGB *)bp;
            for (x = 0; x < w; x++) {
                rgb = *rgbp++;
                DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
            }
        } else if (nc == 2) {
            unsigned char a;
            for (x = 0; x < w; x++) {
                rgb.r = *bp; rgb.g = *bp; rgb.b = *bp++; a = *bp++;
                if (a == 0)
                    rgb = bg;
                else if (a != 255) {
                    png_composite(rgb.r, rgb.r, a, bg.r);
                    png_composite(rgb.g, rgb.g, a, bg.g);
                    png_composite(rgb.b, rgb.b, a, bg.b);
                }
                DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
            }
        } else {
            for (x = 0; x < w; x++) {
                DOPIXEL(*bp, *bp, *bp, 255);
                bp++;
            }
        }
        if (v.pad) dp += v.pad;
    }

end:
    png_read_end(png_ptr, info_ptr);

    if (png_get_text(png_ptr, info_ptr, &tcomments, &(v.ncomments))) {
        if (v.ncomments > MAX_COMMENTS) v.ncomments = MAX_COMMENTS;
        for (i = 0; i < v.ncomments; i++) {
            if (tcomments[i].itxt_length) {
                if (*(tcomments[i].lang_key)) {
                    if (isutf8(tcomments[i].lang_key))
                        v.comments[i].key = strdup(tcomments[i].lang_key);
                    else
                        v.comments[i].key =
                                strdup(win2utf8(tcomments[i].key));
                } else
                    v.comments[i].key =
                            strdup(win2utf8(tcomments[i].key));
                if (isutf8(tcomments[i].text))
                    v.comments[i].text = strdup(tcomments[i].text);
                else
                    v.comments[i].text = strdup("");
            } else {
                v.comments[i].key = strdup(win2utf8(tcomments[i].key));
                v.comments[i].text = strdup(win2utf8(tcomments[i].text));
            }
        }
    }

err:
    if (v.hTransform) {
        cmsDeleteTransform(v.hTransform); v.hTransform = NULL;
    }
    if (v.hInProfile) {
        if (v.iprof) cmsCloseProfile(v.hInProfile); v.hInProfile = NULL;
    }
    if (png_ptr) {
        if (info_ptr)
            png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        else
            png_destroy_read_struct(&png_ptr, NULL, NULL);
    }
    if (buf) free(buf);
    if (row_pointers) free(row_pointers);
    fclose(f);
    return img;
}
