/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "common.h"
#include "../imgview.h"

#define PBM 1
#define PGM 2
#define PPM 3
#define RPBM 4
#define RPGM 5
#define RPPM 6


static void errmsg(char *msg)
{
    message("PNM Error", msg, NULL);
}

static int getmagic(FILE *f)
{
    int ic1, ic2;

    if ((ic1 = getc(f)) == EOF) {
        errmsg("Premature end of file");
        return -1;
    }
    if ((ic2 = getc(f)) == EOF) {
        errmsg("Premature end of file");
        return -1;
    }
    return ic1 * 256 + ic2;
}

static char pgetc(FILE *f)
{
    register int ic;
    register char c;

    if ((ic = getc(f)) == EOF) {
        errmsg("Premature end of file");
        return -1;
    }
    c = (char)ic;

    if (c == '#') {
        do {
            if ((ic = getc(f)) == EOF) {
                errmsg("Premature end of file");
                return -1;
            }
            c = (char)ic;
        } while (c != '\n' && c != '\r');
    }

    return c;
}

static int getbit(FILE *f)
{
    register char c;

    do {
        if ((c = pgetc(f)) == -1)
            return -1;
    } while (c == ' ' || c == '\t' || c == '\n' || c == '\r');

    if (c != '0' && c != '1') {
        errmsg("Junk in file where bits should be");
        return -1;
    }

    return (c == '1' ? 1 : 0);
}

static int getint(FILE *f)
{
    register char c;
    register int i;

    do {
        if ((c = pgetc(f)) == -1)
            return -1;
    } while (c == ' ' || c == '\t' || c == '\n' || c == '\r');

    if (c < '0' || c > '9') {
        errmsg("Junk in file where an integer should be");
        return -1;
    }

    i = 0;
    do {
        i = i * 10 + c - '0';
        if ((c = pgetc(f)) == -1)
            return -1;
    } while (c >= '0' && c <= '9');

    return i;
}

static int readline(FILE *f, unsigned char *buf, int bytes, int width, int fmt)
{
    if (fmt < RPBM) {
        unsigned char *bp = buf;
        int i, c;
        if (fmt == PBM) {
            int bit;
            memset(bp, 0, bytes);
            for (i = 0, bit = 7; i < width; i++, bit--) {
                if ((c = getbit(f)) == -1) return 0;
                if (bit < 0) { bp++; bit = 7; }
                *bp |= (c << bit);
            }
        } else {
            for (i = 0; i < bytes; i++) {
                if ((c = getint(f)) == -1) return 0;
                *(bp++) = (unsigned char)c;
            }
        }
    } else {
        if (fread(buf, 1, bytes, f) < bytes) {
            errmsg("Premature end of file");
            return 0;
        }
    }
    return 1;
}

XImage *readpnm(int fd)
{
    XImage *img = NULL;
    FILE *f;
    int magic;
    int w, h;
    int maxval = 255;
    int fmt;
    unsigned char *line = NULL, *lp;
    int linebytes;
    RGB rgb, *rgbp;
    int x, y;

    f = fdopen(fd, "r");

    if ((magic = getmagic(f)) == -1) goto err;
    fmt = (magic & 0xff) - '0';
    if ((magic >> 8) != 'P' || (fmt < PBM || fmt > RPPM)) {
        errmsg("Unrecognized format");
        goto err;
    }
    if ((w = getint(f)) == -1) goto err;
    if ((h = getint(f)) == -1) goto err;
    if (fmt != PBM && fmt != RPBM) {
        if ((maxval = getint(f)) == -1) goto err;
        sprintf(v.xinfostr, "Maximum Value: %d", maxval);
    }

    v.pad = 0;
    img = mkximage(w, h, fmt == PBM || fmt == RPBM ? 1 : xv.depth, &v.pad);

    linebytes = (img->depth == 1 ? (w + 7) / 8 :
            w * ((fmt == PPM || fmt == RPPM) ? 3 : 1));

    if (fmt == RPBM) {
        int imgbytes = linebytes * h;
        if (fread(img->data, 1, imgbytes, f) < imgbytes) {
            errmsg("Premature end of file");
            rmximage(img); img = NULL; goto err;
        }
        goto end;
    }

    if (img->depth != 1 && maxval < 255)
        mkrgbtable(p_red, p_green, p_blue, maxval);

    line = (unsigned char *)malloc(linebytes);

    dp = (unsigned char *)img->data;
    idp = (unsigned int *)dp;

    for (y = 0; y < h; y++) {
        if (! readline(f, line, linebytes, w, fmt))
                { rmximage(img); img = NULL; goto err; }
        lp = line;
        if (img->depth == 1) {
            memcpy(dp, lp, linebytes); dp += linebytes;
        } else if (fmt == PPM || fmt == RPPM) {
            rgbp = (RGB *)lp;
            for (x = 0; x < w; x++) {
                rgb = *rgbp++;
                DOPIXEL(rgb.r, rgb.g, rgb.b, maxval);
            }
        } else {
            for (x = 0; x < w; x++) {
                DOPIXEL(*lp, *lp, *lp, maxval);
                lp++;
            }
        }
        if (v.pad) dp += v.pad;
    }

end:
err:
    if (line) free(line);
    fclose(f);
    return img;
}
